#include <Wire.h>
#include <SoftSerial.h>
#include <EEPROM.h>
#include <TM1637Display.h>

float average[2];
int nbr_slot=15;
int value[16];  // array of (nbr_slot + 1) 
float std_dev;
int samples=10;
int threshold=105; 
int monster_threshold=200; // a bit high?
unsigned long watchdog_time;
// pin connections
#define DIO 6
#define CLK 7
int RAZ_pin = 8;
int volume_pin = A9;
int watchdog_pin = 10;
int power_pin=A12;
// pin 0 and 2 go to dfplayer RX TX
//MP3 Settings
 SoftSerial mySerial(0, 2); // if 1 is used it turns the on board led
# define Start_Byte 0x7E
# define Version_Byte 0xFF
# define Command_Length 0x06
# define End_Byte 0xEF
# define Acknowledge 0x00 //Returns info with command 0x41 [0x01: info, 0x00: no info]

int score =0;
int volume,hit,power_read,power_recheck;
bool isNewHit = true;
bool enable=false;
long hitTime;

TM1637Display display = TM1637Display(CLK, DIO);


void setup() {
  display.setBrightness(2);
  //initialize serial link to mp3 player and wait until it is ready
  mySerial.begin(9600);
  while(!mySerial.available()) delay(1);
  //volume is adjusted only at sartup
  volume=map(analogRead(volume_pin),0,1023,0,30);
  delay(300); //needed with updated micronucleus
  setVolume(volume);
  celebrate_a_kill(3); // get ready to the next fight
  pinMode(RAZ_pin, INPUT_PULLUP); 
  //if RAZ button is pushed at startup, saved score is reset to zero 
  if ( digitalRead(RAZ_pin) == 0 ) RAZ();
  //read score from byte 0 and 1
  score = (EEPROM.read(0) << 8) + EEPROM.read(1);
  //turn 4 digits display on and display saved score  
   //score = 0;
   display.showNumberDec(score);
   //the following 3 lines maybe needed only if your charging circuit has an automatic standby mode
   watchdog_time=millis();
   pinMode(watchdog_pin,OUTPUT);
   digitalWrite(watchdog_pin, HIGH);
   }




void loop() {

// process_average sums the last value of power 
// process_std_dev process the square of the standard deviation and is above some level considere it is a kill
// some more checks are performed to exclude phases where the swatter button is pushed or release (in this case standard deviation goes extremlt high
// if hit is detected play sound
// if score multiple of 10 play bonus
// boolean isNewHit makes sure we are not recounting the current hit
// boolean enable is true when the swatter button is pushed

hit=0;
if ( !isNewHit && (millis() - hitTime) > 500 ) isNewHit=true ;
process_average();
process_std_dev();


if (isNewHit && enable) {
if ((hit>=100)  ){
        //we got a kill, let's play something
        hitTime=millis();
        isNewHit=false;
       // uncomment for debug
       // display.showNumberDec(int(std_dev));
        if (hit==200){
        //this is a monster kill
          celebrate_a_kill(1);
          
        }
        else {
        celebrate_a_kill(0);
        
        }
       
        EEPROM.write(0, ++score >> 8); //write first byte of score
        EEPROM.write(1, score & 0xFF); //write second byte of score
      //comment for debug
       display.showNumberDec(score);
        
        if (((score)%10)==0) {
          // this is a bonus - level up!
          //wait a bit before playing again
          delay(2000);
          celebrate_a_kill(2);
        }
      }
}

// when debugging works ok but when serial write not done detection is not good
// insert delay as not writing make code run faster and my disturb the tuning...
// after hours of tuning...not sure we need this delay anymore but 
delay(10);
watchdog();
//for debug
if ( digitalRead(RAZ_pin) == 0 ) display.showNumberDec(int(std_dev));
 
}





void process_average()
{

//shift value[] 
for (int i=nbr_slot; i>0 ; i--) value[i]=value[i-1];


value[0]=0;
 for (int i=0; i < samples; i++) { // pin is read "samples" times 
    power_read=analogRead(power_pin);
    value[0] = value[0] + power_read;
   }
 value[0] = value[0]/samples;
 //
 // sliding average calc (in fact the sum of the last "nbr_slot" of measures
 // this is not the average but  nbr_slot * the average
 // it simplfies calculation and save compute cycles
 average[1]=average[0]; // we keep the last average
 average[0] = average[0] + value[0] - value[nbr_slot];
 if (average[0] < 700*nbr_slot) {
    enable=false;
    //display.showNumberDec(8000);
 }
 
}

void process_std_dev() //calculate the square of standard deviation
{

   //calculate only if average > 900 for two cycles
   if ((average[0] > 900*nbr_slot) && (average[1] > 900*nbr_slot)) {
   for (int i=0; i < nbr_slot; i++) {
      std_dev= std_dev+((value[i]-(average[0]/nbr_slot))*(value[i]-(average[0]/nbr_slot)));
    }
   std_dev= std_dev/nbr_slot;
   power_recheck = analogRead(power_pin) ;
   // do we have a hit ? 
   // 1000 could be lower
   // 200 could be higher to avoid false detection when button is released
   if (std_dev >1000) std_dev=0; //we are in trasition phase this is not a mosquito
   //initial values 80 and 100 + changed from 200 to 600 for transition to low exclusion
   if ((std_dev >threshold) && (value[0]  > 700 ) && (power_recheck > 700)) hit=100; //we got a small hit only if power is on or we could do a bad detection 
   if ((std_dev >monster_threshold)&& (value[0]  > 700 ) && (power_recheck > 700))  hit=200; //we got a monster kill only if power is on or we could do a bad detection
   if (std_dev <40) {
    enable=true; //std_dev is stabilized after transition, we can start counting
    //display.showNumberDec(1000);
   }
    
 }
}

void watchdog()
{
  if ( (millis() - watchdog_time) > 28000 ) {
    digitalWrite(watchdog_pin, LOW);
    delay(40);
    digitalWrite(watchdog_pin, HIGH); 
    watchdog_time=millis();
  }
}

void RAZ()
{
  //Two bytes are needed to store the score
  // we reset the score to 0
  EEPROM.write(0, 0);
  EEPROM.write(1, 0);
}

void celebrate_a_kill(int sound_type)
{
//wait for serial
  while(!mySerial.available()){
    delay(1);
  }
  // 0 is random
  // 1 is monster kill = track 1 (listed by inode on card (ls -id)
  // 2 is level up
  // 3 is power on
  if (sound_type==0)   setTrack(int(random(4,9))); //SD card contains 9 files
  if (sound_type==1)   setTrack(1);
  if (sound_type==2)   setTrack(2);
  if (sound_type==3)   setTrack(3);
  
}

// following functions are for the dfplayer
//thanks to https://www.dfrobot.com/blog-1422.html

void setTrack(int track) //select sound to play
{
  execute_CMD(0x03, 0, track); // Set the track (0x00~0x30)
  //delay(100); // needed?
}

void execute_CMD(byte CMD, byte Par1, byte Par2)
// Excecute the command and parameters for dfplayer communication
{
  // Calculate the checksum (2 bytes)
  word checksum = -(Version_Byte + Command_Length + CMD + Acknowledge + Par1 + Par2);
  // Build the command line
  byte Command_line[10] = { Start_Byte, Version_Byte, Command_Length, CMD, Acknowledge,
                            Par1, Par2, highByte(checksum), lowByte(checksum), End_Byte
                          };
  //Send the command line to the module
  for (byte k = 0; k < 10; k++)
  {
     mySerial.write( Command_line[k]);
    
  }
}

void setVolume(int volume)
{
  execute_CMD(0x06, 0, volume); // Set the volume (0x00~0x30)
  delay(200);
}
